import os

from dotenv import load_dotenv
from slack_sdk import WebClient
from slack_sdk.errors import SlackApiError

load_dotenv()

# Load environment variables
production = os.getenv("PRODUCTION")
slack_token = os.getenv("SLACK_ACCESS_TOKEN")


class Slack:
    """
    A class to send messages to a Slack channel
    attributes: slack_workspace_token
    """

    def __init__(self):
        self.slack_workspace_token = slack_token

    ########################################################
    # Define the send_message function
    ########################################################
    def send_message(self, channel, message, username):
        """
        Send a message to the Slack channel
        :param channel:
        :param message: Message to send
        :param username: Username to send the message as
        If the slack_workspace_token is not set, print the message instead of sending it to Slack
        """
        # Create a WebClient object

        if not self.slack_workspace_token:
            print(message)
            return

        client = WebClient(token=self.slack_workspace_token)

        try:
            response = client.chat_postMessage(
                channel=channel, text=f"{message}", username=username
            )
            assert response["message"]["text"] == f"{message}"
        except SlackApiError as e:
            # You will get a SlackApiError if "ok" is False
            assert e.response["ok"] is False
            assert e.response["error"]  # str like 'invalid_auth', 'channel_not_found'
            print(f"Got an error: {e.response['error']}")
            # Also receive a corresponding status_code
            assert isinstance(e.response.status_code, int)
            print(f"Received a response status_code: {e.response.status_code}")



import logging

from py_alpaca_api import Stock, Trading

from alpaca_daily_losers.global_functions import (
    get_ticker_data,
    send_message,
    send_position_messages,
)


class ClosePositions:
    def __init__(self, trading_client: Trading, stock_client: Stock, py_logger: logging.Logger):
        self.trade = trading_client
        self.stock = stock_client
        self.py_logger = py_logger

    def sell_positions_from_criteria(self) -> None:
        """
        Sells positions based on the defined sell criteria, including RSI and
        Bollinger Band High Index (BBHI) thresholds, as well as take profit
        and stop loss conditions.

        This method first retrieves the current positions, filters out the
        cash position, and then uses the `get_stocks_to_sell()` method to
        determine which positions should be sold. It then calls
        the `_sell_positions()` method to execute the sell orders and
        sends messages to notify of the sold positions.

        If no positions meet the sell criteria, a message is sent
        indicating that no sell opportunities were found.

        Raises:
            Exception: If an error occurs while selling the positions.
        """

        # console.print("Selling positions based on criteria...", style="bold green")
        try:
            stocks_to_sell = self.get_stocks_to_sell()
            if not stocks_to_sell:
                send_message("No sell opportunities found.")
                return
            current_positions = self.trade.positions.get_all()
            sold_positions = self._sell_positions(stocks_to_sell, current_positions)
            send_position_messages(sold_positions, "sell")
        except Exception as e:
            self.py_logger.error(f"Error selling positions from criteria. Error: {e}")

    def _sell_positions(self, stocks_to_sell, current_positions) -> list:
        """
        Sell positions for the given stocks.

        Args:
            stocks_to_sell (list): List of symbols for the stocks to sell.
            current_positions (pandas.DataFrame): DataFrame containing
            the current positions.

        Returns:
            list: List of dictionaries representing the sold positions,
            each containing the symbol and quantity.
        """
        sold_positions = []
        for symbol in stocks_to_sell:
            try:
                qty = current_positions[current_positions["symbol"] == symbol]["qty"].values[0]
                self.trade.positions.close(symbol_or_id=symbol, qty=qty)
                sold_positions.append({"symbol": symbol, "qty": qty})
            except Exception as e:
                self.py_logger.warning(f"Could not close {symbol}. Error: {e}")
                send_message(f"Error selling {symbol}: {e}")
        return sold_positions

    def get_stocks_to_sell(self) -> list:
        """
        Retrieves a list of stocks to sell based on specific criteria.

        Returns:
            A list of stocks to sell.
        """

        current_positions = self.trade.positions.get_all()
        non_cash_positions = current_positions[current_positions["symbol"] != "Cash"]

        if non_cash_positions.empty:
            return []

        current_positions_symbols = non_cash_positions["symbol"].tolist()
        assets_history = get_ticker_data(current_positions_symbols, self.stock, self.py_logger)

        RSI_COLUMNS = ["rsi14", "rsi30", "rsi50", "rsi200"]
        BBHI_COLUMNS = ["bbhi14", "bbhi30", "bbhi50", "bbhi200"]

        criterion1 = assets_history[RSI_COLUMNS] >= 70
        criterion2 = assets_history[BBHI_COLUMNS] == 1
        sell_criteria = criterion1.any(axis=1) | criterion2.any(axis=1)

        sell_filtered_df = assets_history[sell_criteria]
        stocks_to_sell = sell_filtered_df["symbol"].tolist()

        take_profit_list = non_cash_positions[non_cash_positions["profit_pct"] > 8.0][
            "symbol"
        ].tolist()
        stop_loss_list = non_cash_positions[non_cash_positions["profit_pct"] < -8.0][
            "symbol"
        ].tolist()

        for take_profit, stop_loss in zip(take_profit_list, stop_loss_list):
            if take_profit not in stocks_to_sell:
                stocks_to_sell.append(take_profit)
            if stop_loss not in stocks_to_sell:
                stocks_to_sell.append(stop_loss)

        return stocks_to_sell



import os

from dotenv import load_dotenv
from openai import OpenAI

load_dotenv()


class OpenAIAPI:
    def __init__(self):
        self.api_key = os.getenv("OPENAI_API_KEY")

    ########################################################
    # Define the OpenAi chat function
    ########################################################
    def chat(self, msgs):
        """
        Chat with the OpenAI API
        :param msgs: List of messages
        return: OpenAI response
        """
        openai = OpenAI(api_key=self.api_key)
        response = openai.chat.completions.create(model="gpt-3.5-turbo", messages=msgs)
        message = response
        return message

    ########################################################
    # Define the get_market_sentiment function
    ########################################################
    def get_sentiment_analysis(self, title, symbol, article):
        """
        get_sentiment_analysis(self, title, symbol, article)

        This method is used to get the sentiment analysis for financial news.
        It takes in the title, symbol, and article as parameters.

        Parameters:
            - title (str): The title of the news.
            - symbol (str): The stock symbol associated with the news.
            - article (str): The content of the news article.

        Returns:
            - signal (str): The sentiment analysis result as either
            "BEARISH", "BULLISH", or "NEUTRAL".

        """
        message_history = []
        sentiments = []
        # Send the system message to the OpenAI API
        system_message = "You will work as a Sentiment Analysis for Financial news. \
                        I will share news headline, stock symbol and article. \
                        You will only answer as:\n\n BEARISH,BULLISH,NEUTRAL. \
                        No further explanation. \n Got it?"

        message_history.append({"content": system_message, "role": "user"})
        self.chat(message_history)

        # Send the article to the OpenAI API
        user_message = "{}\n{}\n{}".format(title, symbol, article)

        message_history.append({"content": user_message, "role": "user"})
        response = self.chat(message_history)
        sentiments.append(
            {
                "title": title,
                "symbol": symbol,
                "article": article,
                "signal": response.choices[0].message.content,
            }
        )
        message_history.pop()
        # Return the sentiment
        return sentiments[0]["signal"]



import logging

import pandas as pd
from py_alpaca_api.trading import Trading

from alpaca_daily_losers.global_functions import send_message, send_position_messages


class Liquidate:
    def __init__(self, trading_client: Trading, py_logger: logging.Logger):
        self.trade = trading_client
        self.py_logger = py_logger

    @staticmethod
    def calculate_cash_needed(total_holdings: float, cash_row: pd.DataFrame) -> float:
        """
        Calculate the amount of cash needed to liquidate a portion of holdings.

        Parameters:
        total_holdings (float): The total value of the holdings to be liquidated.
        cash_row (pd.DataFrame): A DataFrame containing the cash information.

        Returns:
        float: The amount of cash needed for liquidation, including a fixed fee of $5.00.
        """
        return (total_holdings * 0.1 - cash_row["market_value"].iloc[0]) + 5.00

    @staticmethod
    def get_top_performers(current_positions: pd.DataFrame) -> pd.DataFrame:
        """
        Returns the top performers from the given current positions DataFrame.

        Parameters:
        - current_positions (pd.DataFrame): DataFrame containing the current positions.

        Returns:
        - pd.DataFrame: DataFrame containing the top performers.
        """
        non_cash_positions = current_positions[current_positions["symbol"] != "Cash"].sort_values(
            by="profit_pct", ascending=False
        )
        return non_cash_positions.iloc[: int(len(non_cash_positions) // 2)]

    def liquidate_positions(self) -> None:
        """
        Liquidates positions to make cash 10% of the portfolio.

        This method sells positions in order to meet the requirement of having cash
        equal to 10% of the portfolio's total value. It identifies the top performers
        in the current positions and calculates the amount of cash needed to meet the
        requirement. It then sells the necessary amount of shares for each top performer.

        Returns:
            None
        """
        current_positions = self.trade.positions.get_all()
        if current_positions[current_positions["symbol"] != "Cash"].empty:
            self.send_liquidation_message("No positions available to liquidate for capital")
            return
        cash_row = current_positions[current_positions["symbol"] == "Cash"]
        total_holdings = current_positions["market_value"].sum()
        sold_positions = []
        if cash_row["market_value"].iloc[0] / total_holdings < 0.1:
            top_performers = self.get_top_performers(current_positions)
            top_performers_market_value = top_performers["market_value"].sum()
            cash_needed = self.calculate_cash_needed(total_holdings, cash_row)

            for index, row in top_performers.iterrows():
                amount_to_sell = int(
                    (row["market_value"] / top_performers_market_value) * cash_needed
                )
                if amount_to_sell == 0:
                    continue
                try:
                    self.trade.orders.market(
                        symbol=row["symbol"],
                        notional=amount_to_sell,
                        side="sell",
                    )
                except Exception as e:
                    self.py_logger.warning(
                        f"Error liquidating position {row['symbol']}. Error: {e}"
                    )
                    self.send_liquidation_message(f"Error selling {row['symbol']}: {e}")
                    continue
                else:
                    sold_positions.append(
                        {
                            "symbol": row["symbol"],
                            "notional": round(amount_to_sell, 2),
                        }
                    )
        send_position_messages(sold_positions, "liquidate")

    @staticmethod
    def send_liquidation_message(message: str):  # Renamed method to be more specific
        send_message(message)



import logging
import os

import pandas as pd
from dotenv import load_dotenv
from py_alpaca_api import PyAlpacaAPI

from alpaca_daily_losers.close_positions import ClosePositions
from alpaca_daily_losers.global_functions import (
    get_ticker_data,
    send_message,
    send_position_messages,
)
from alpaca_daily_losers.liquidate import Liquidate
from alpaca_daily_losers.openai import OpenAIAPI
from alpaca_daily_losers.statistics import Statistics

# get a custom logger & set the logging level
py_logger = logging.getLogger(__name__)
logging.basicConfig(
    level=logging.WARNING,
    format="%(name)s %(asctime)s %(levelname)s %(message)s",
)

load_dotenv()

api_key = str(os.getenv("API_KEY"))
api_secret = str(os.getenv("API_SECRET"))
api_paper = True if os.getenv("API_PAPER") == "True" else False


class DailyLosers:
    def __init__(self):
        self.alpaca = PyAlpacaAPI(api_key=api_key, api_secret=api_secret, api_paper=api_paper)
        self.production = True if os.getenv("PRODUCTION") == "True" else False
        self.liquidate = Liquidate(trading_client=self.alpaca.trading, py_logger=py_logger)
        self.close = ClosePositions(
            trading_client=self.alpaca.trading,
            stock_client=self.alpaca.stock,
            py_logger=py_logger,
        )
        self.statistics = Statistics(account=self.alpaca.trading.account, py_logger=py_logger)

    def run(self):
        """
        Executes the main logic of the program.

        This method performs the following steps:
        1. Attempts to sell positions based on certain criteria.
        2. Attempts to liquidate positions for capital.
        3. Checks for new buy opportunities.

        If any of the steps encounter an error, it logs the error and
        continues to the next step.
        """
        try:
            self.close.sell_positions_from_criteria()
        except Exception as e:
            py_logger.error(f"Error selling positions from criteria. Error {e}")
            pass
        try:
            self.liquidate.liquidate_positions()
        except Exception as e:
            py_logger.error(f"Error liquidating positions for capital. Error: {e}")
            pass
        try:
            self.check_for_buy_opportunities()
        except Exception as e:
            py_logger.error(f"Error entering new positions. Error {e}")

    ########################################################
    # Define the check_for_buy_opportunities method
    ########################################################
    def check_for_buy_opportunities(self) -> None:
        """
        Checks for buy opportunities based on daily losers and opens positions if any are found.

        Returns:
            None
        """
        losers = self.get_daily_losers()
        tickers = self.filter_tickers_with_news(losers)

        if len(tickers) > 0:
            print(f"Found [bold]{len(tickers)}[/bold] buy opportunities.")
            self.open_positions(tickers=tickers)
        else:
            print("No buy opportunities found")

    ########################################################
    # Define the open_positions method
    ########################################################
    def open_positions(self, tickers: list, ticker_limit: int = 8) -> None:
        """
        Opens buying orders based on buy opportunities and openai sentiment.
        Limits the number of stocks to buy to 8 by default.

        Args:
            tickers (list): A list of ticker symbols to buy.
            ticker_limit (int, optional): The maximum number of tickers to buy. Defaults to 8.

        Returns:
            None
        """

        available_cash = self.alpaca.trading.account.get().cash

        if len(tickers) == 0:
            send_message("No tickers to buy.")
            return
        else:
            notional = (available_cash / len(tickers[:ticker_limit])) - 1

        bought_positions = []

        for ticker in tickers[:ticker_limit]:
            try:
                self.alpaca.trading.orders.market(symbol=ticker, notional=notional)
            except Exception as e:
                py_logger.warning(f"Error entering new position for {ticker}. Error: {e}")
                send_message(message=f"Error buying {ticker}: {e}")
                continue
            else:
                bought_positions.append({"symbol": ticker, "notional": round(notional, 2)})

        send_position_messages(positions=bought_positions, pos_type="buy")

    ########################################################
    # Define the update_or_create_watchlist method
    ########################################################
    def update_or_create_watchlist(self, name, symbols) -> None:
        """
        Updates an existing watchlist with the given name and symbols,
        or creates a new watchlist if it doesn't exist.

        Args:
            name (str): The name of the watchlist.
            symbols (List[str]): A list of symbols to be added to the watchlist.

        Returns:
            None

        Raises:
            Exception: If the watchlist cannot be updated or created.

        """
        try:
            self.alpaca.trading.watchlists.update(watchlist_name=name, symbols=symbols)
        except Exception as e:
            py_logger.info(
                f"Watchlist could not be updated: {e}:\nTrying to create new watchlist with \
                    name: {name}"
            )
            try:
                self.alpaca.trading.watchlists.create(name=name, symbols=symbols)
            except Exception as e:
                py_logger.error(f"Could not create or update the watchlist {name}.\nError: {e}")

    ########################################################
    # Define the filter_tickers_with_news method
    ########################################################
    def filter_tickers_with_news(self, tickers) -> list:
        """
        Filters a list of tickers based on news sentiment analysis.

        Args:
            tickers (list): A list of ticker symbols.

        Returns:
            list: A list of tickers that have positive news sentiment.
        """
        openai = OpenAIAPI()
        filtered_tickers = []

        for i, ticker in enumerate(tickers):
            try:
                articles = self.alpaca.trading.news.get_news(symbol=ticker)
            except Exception as e:
                py_logger.warning(f"Error getting articles for {ticker}. Error {e}")
                continue

            if articles is None:
                continue

            if len(articles) > 0:
                bullish = 0
                bearish = 0
                for art in articles[:3]:
                    sentiment = openai.get_sentiment_analysis(
                        title=art["title"],
                        symbol=art["symbol"],
                        article=art["content"],
                    )
                    if sentiment == "BULLISH":
                        bullish += 1
                    else:
                        bearish += 1

                if bullish > bearish:
                    filtered_tickers.append(ticker)

        if len(filtered_tickers) == 0:
            print("No tickers with news found")
            return []

        self.update_or_create_watchlist(name="DailyLosers", symbols=filtered_tickers)

        return self.alpaca.trading.watchlists.get_assets(watchlist_name="DailyLosers")

    ########################################################
    # Define the get_daily_losers method
    ########################################################
    def get_daily_losers(self) -> list:
        """
        Retrieves the daily losers from Alpaca stock predictor,
        filters them based on buy criteria,
        and updates the watchlist with the filtered losers.

        Returns:
            A list of assets in the DailyLosers watchlist.
        """
        losers = self.alpaca.stock.predictor.get_losers_to_gainers()
        losers = get_ticker_data(
            tickers=losers, stock_client=self.alpaca.stock, py_logger=py_logger
        )
        losers = self.buy_criteria(losers)

        if len(losers) == 0:
            send_message("No daily losers found.")
            return []

        for i, ticker in enumerate(losers):
            try:
                sentiment = self.alpaca.trading.recommendations.get_sentiment(ticker)
            except Exception as e:
                py_logger.info(f"Error getting sentiment from Yahoo. Error: {e}")
                sentiment = "NEUTRAL"

            if sentiment == "NEUTRAL" or sentiment == "BEARISH":
                losers.remove(ticker)

        self.update_or_create_watchlist(name="DailyLosers", symbols=losers)

        return self.alpaca.trading.watchlists.get_assets(watchlist_name="DailyLosers")

    ########################################################
    # Define the buy_criteria method
    ########################################################
    def buy_criteria(self, data: pd.DataFrame) -> list:
        """
        Applies buy criteria to the given DataFrame and returns a
        list of symbols that meet the criteria.

        Parameters:
        - data (pd.DataFrame): The DataFrame containing the data
        to apply the buy criteria on.

        Returns:
        - list: A list of symbols that meet the buy criteria.

        Raises:
        - None

        Example usage:
        ```
        data = pd.DataFrame(...)
        symbols = buy_criteria(data)
        print(symbols)
        ```

        Note:
        - The buy criteria is based on two conditions:
            1. RSI (Relative Strength Index) values in the columns
            'rsi14', 'rsi30', 'rsi50', 'rsi200' should be less than or equal to 30.
            2. BBLO (Bollinger Bands Lower Indicator) values in the
            columns 'bblo14', 'bblo30', 'bblo50', 'bblo200' should be equal to 1.
        - If no symbols meet the buy criteria, an empty list is returned.
        - The watchlist named 'DailyLosers' is updated or created with
        the symbols that meet the buy criteria.
        - The assets in the 'DailyLosers' watchlist are returned.

        """

        RSI_COLUMNS = ["rsi14", "rsi30", "rsi50", "rsi200"]
        BBLO_COLUMNS = ["bblo14", "bblo30", "bblo50", "bblo200"]

        criterion1 = data[RSI_COLUMNS] <= 30
        criterion2 = data[BBLO_COLUMNS] == 1
        buy_criteria = criterion1.any(axis=1) | criterion2.any(axis=1)

        buy_filtered_data = data[buy_criteria]

        filtered_data = list(buy_filtered_data["symbol"])

        if len(filtered_data) == 0:
            print("No tickers meet the buy criteria")
            return []

        self.update_or_create_watchlist(name="DailyLosers", symbols=filtered_data)

        return self.alpaca.trading.watchlists.get_assets(watchlist_name="DailyLosers")



import logging

from py_alpaca_api.trading.account import Account


class Statistics:
    def __init__(self, account: Account, py_logger: logging.Logger) -> None:
        self.account = account
        self.py_logger = py_logger



import os
from datetime import datetime, timedelta

import pandas as pd
from dotenv import load_dotenv
from py_alpaca_api import Stock
from pytz import timezone
from ta.momentum import RSIIndicator
from ta.volatility import BollingerBands

from alpaca_daily_losers.slack import Slack

load_dotenv()

tz = timezone("US/Eastern")
ctime = datetime.now(tz)
today = ctime.strftime("%Y-%m-%d")
previous_day = (ctime - timedelta(days=1)).strftime("%Y-%m-%d")
year_ago = (ctime - timedelta(days=365)).strftime("%Y-%m-%d")

production = os.getenv("PRODUCTION")
slack_username = os.getenv("SLACK_USERNAME")


########################################################
# Define the get_ticker_data method
########################################################
def get_ticker_data(tickers, stock_client: Stock, py_logger) -> pd.DataFrame:
    df_tech = []

    for i, ticker in enumerate(tickers):
        try:
            history = stock_client.history.get_stock_data(
                symbol=ticker, start=year_ago, end=previous_day
            )
        except Exception as e:
            py_logger.warning(f"Error get historical data for {ticker}. Error: {e}")
            continue

        try:
            for n in [14, 30, 50, 200]:
                history["rsi" + str(n)] = RSIIndicator(close=history["close"], window=n).rsi()
                history["bbhi" + str(n)] = BollingerBands(
                    close=history["close"], window=n, window_dev=2
                ).bollinger_hband_indicator()
                history["bblo" + str(n)] = BollingerBands(
                    close=history["close"], window=n, window_dev=2
                ).bollinger_lband_indicator()
            df_tech_temp = history.tail(1)
            df_tech.append(df_tech_temp)
        except KeyError:
            pass

    if df_tech:
        df_tech = [x for x in df_tech if not x.empty]
        df_tech = pd.concat(df_tech)
    else:
        df_tech = pd.DataFrame()

    return df_tech


########################################################
# Define the _send_position_messages method
########################################################
def send_position_messages(positions: list, pos_type: str):
    """
    Sends position messages based on the type of position.
    Args:
        positions (list): List of position dictionaries.
        pos_type (str): Type of position ("buy", "sell", or "liquidate").
    Returns:
        bool: True if message was sent successfully, False otherwise.
    """
    position_names = {
        "sell": "sold",
        "buy": "bought",
        "liquidate": "liquidated",
    }

    try:
        position_name = position_names[pos_type]
    except KeyError:
        raise ValueError('Invalid type. Must be "sell", "buy", or "liquidate".')

    if not positions:
        position_message = f"No positions to {pos_type}"
    else:
        position_message = f"Successfully {position_name} the following positions:\n"

        for position in positions:
            if position_name == "liquidated":
                qty_key = "notional"
            elif position_name == "sold":
                qty_key = "qty"
            else:
                qty_key = "notional"

            qty = position[qty_key]
            symbol = position["symbol"]

            position_message += f"{qty} shares of {symbol}\n"
    return send_message(position_message)


def send_message(message):
    """
    Send a message to Slack
    :param message: str: message to send
    """
    slack = Slack()
    if production == "False":
        print(f"Message: {message}")
    else:
        slack.send_message(channel="#app-development", message=message, username=slack_username)
